package types;

// The expressions in our mini-language.
public interface Expression {

    // Belongs to the visitor pattern, see the `ExpressionVisitor` interface.
    public <T, E extends Exception> T accept(ExpressionVisitor<T, E> v) throws E;

    
    // Here follow all the possible expressions:

    public static final class True implements Expression {
        @Override
        public String toString() {
            return "true";
        }

        @Override
        public <T, E extends Exception> T accept(ExpressionVisitor<T, E> v) throws E {
            return v.visit(this);
        }
    }

    public static final class False implements Expression {
        @Override
        public String toString() {
            return "false";
        }

        @Override
        public <T, E extends Exception> T accept(ExpressionVisitor<T, E> v) throws E {
            return v.visit(this);
        }
    }

    public static final class Int implements Expression {
        private final int value;

        public Int(int value) {
            // We restrict integer expressions to only positive values.
            assert value >= 0;
            this.value = value;
        }

        @Override
        public String toString() {
            return String.valueOf(value);
        }

        @Override
        public <T, E extends Exception> T accept(ExpressionVisitor<T, E> v) throws E {
            return v.visit(this);
        }
    }

    public static final class Neg implements Expression {
        public final Expression e;

        public Neg(Expression e) {
            this.e = e;
        }

        @Override
        public String toString() {
            return "- (" + e + ")";
        }

        @Override
        public <T, E extends Exception> T accept(ExpressionVisitor<T, E> v) throws E {
            return v.visit(this);
        }
    }

    public static final class Add implements Expression {
        public final Expression e1, e2;

        public Add(Expression e1, Expression e2) {
            this.e1 = e1;
            this.e2 = e2;
        }

        @Override
        public String toString() {
            return "(" + e1 + ") + (" + e2 + ")";
        }

        @Override
        public <T, E extends Exception> T accept(ExpressionVisitor<T, E> v) throws E {
            return v.visit(this);
        }
    }

    public static final class Or implements Expression {
        public final Expression e1, e2;

        public Or(Expression e1, Expression e2) {
            this.e1 = e1;
            this.e2 = e2;
        }

        @Override
        public String toString() {
            return "(" + e1 + ") || (" + e2 + ")";
        }

        @Override
        public <T, E extends Exception> T accept(ExpressionVisitor<T, E> v) throws E {
            return v.visit(this);
        }
    }

    public static final class Equals implements Expression {
        public final Expression e1, e2;

        public Equals(Expression e1, Expression e2) {
            this.e1 = e1;
            this.e2 = e2;
        }

        @Override
        public String toString() {
            return "(" + e1 + ") == (" + e2 + ")";
        }

        @Override
        public <T, E extends Exception> T accept(ExpressionVisitor<T, E> v) throws E {
            return v.visit(this);
        }
    }

    public static final class IfThenElse implements Expression {
        public final Expression cond, then, else_;

        public IfThenElse(Expression cond, Expression then, Expression else_) {
            this.cond = cond;
            this.then = then;
            this.else_ = else_;
        }

        @Override
        public String toString() {
            return "if (" + cond + ") then (" + then + ") else (" + else_ + ")";
        }

        @Override
        public <T, E extends Exception> T accept(ExpressionVisitor<T, E> v) throws E {
            return v.visit(this);
        }
    }


    // We also define a bunch of convenience methods for creating expressions:

    public static final Expression true_ = new True();

    public static final Expression false_ = new False();

    public static Expression int_(int value) {
        return new Int(value);
    }

    public static Expression neg(Expression e) {
        return new Neg(e);
    }

    public static Expression add(Expression e1, Expression e2) {
        return new Add(e1, e2);
    }

    public static Expression or(Expression e1, Expression e2) {
        return new Or(e1, e2);
    }

    public static Expression eq(Expression e1, Expression e2) {
        return new Equals(e1, e2);
    }

    public static Expression ite(Expression cond, Expression then, Expression else_) {
        return new IfThenElse(cond, then, else_);
    }

}
